@@ -1,12 +1,12 @@
-from .Base import _Base
+from dynamofl.api.DatasourceAPI import DatasourceAPI
 
-class _Datasource(_Base):
-    def __init__(self, dfl, key, type):
-        super().__init__(dfl.token, dfl.host)
-        self.dfl = dfl
+
+class _Datasource:
+    def __init__(self, key, type, api: DatasourceAPI):
+        self._api = api
         self.key: str = key
         self.trainers = {}
-        
+
         # Can be 'horizontal', 'label' or 'feature'
         self.type = type
 
@@ -19,11 +19,15 @@ class _Datasource(_Base):
         if self.type is not None and self.type != 'horizontal':
             params['type'] = self.type
 
-        self._make_request('POST', f'/datasources/{self.key}/trainers', params=params)
+        self._api.create_trainers(self.key, params=params)
+
         self.trainers[key] = {
             'train': train_callback,
             'test': test_callback,
         }
 
         if model_path is not None:
-            self.trainers[key]['model_path'] = model_path
\ No newline at end of file
+            self.trainers[key]['model_path'] = model_path
+
+    def get_datasource(self):
+        return self._api.get_datasource(key=self.key)
